#!/usr/bin/python
# -*- coding: utf-8 -*-

import json


class OpenApi:
    version = '0.1'

    def __init__(self, connection):
        self.connection = connection

    def execute(self, method, **params):
        response = getattr(self.connection, method)(params)
        if response['status'] != 0:
            self.txn_id = None
            raise Exception('Method {0} returned non-zero status {1} and\
                             error {2}'.format(method, response['status'], response['error_message']))
        else:
            return response.get('result', None)


class ActivationParameter:
    version = '0.1'

    def __init__(self, var_name, var_value):
        self.resource_type_id = 0
        self.service_template_id = 0

        self.var_name = var_name
        self.var_value = var_value

    @classmethod
    def from_raw(cls, raw):
        return cls(raw['var_name'], raw['var_value'])

    def set_resource_type_id(self, resource_type_id):
        self.resource_type_id = resource_type_id

    def set_service_template_id(self, service_template_id):
        self.service_template_id = service_template_id

    def save(self, db_connection):
        db_connection.save_activation_parameter(self)


class ResourceType:
    version = '0.1'

    def __init__(self, resource_type_id, name, resclass_name, measurement_type, limit):
        self.resource_type_id = resource_type_id
        self.name = name
        self.resclass_name = resclass_name
        self.measurement_type = measurement_type
        self.limit = limit
        # attributes and act_params only returned from service template
        self.attributes = []
        self.act_params = []

    @classmethod
    def from_raw(cls, raw):
        if 'measurement_type' in raw:
            measurement_type = raw['measurement_type']
        else:
            measurement_type = raw['measurement_unit']

        return cls(raw['resource_type_id'], raw['name'], raw['resclass_name'],
                   measurement_type, raw['limit'])

    def set_attributes(self, attributes):
        self.attributes = attributes

    def set_act_params(self, act_params):
        self.act_params = act_params

    def save(self, db_connection):
        db_connection.save_resource_type(self)


class ServiceTemplate:
    version = '0.1'

    def __init__(self, st_id, name, is_active, version):
        self.st_id = st_id
        self.name = name
        self.resource_types = []
        self.resource_model = ''
        self.active = is_active
        self.version = version

    @classmethod
    def from_raw(cls, raw):
        return cls(raw['st_id'], raw['name'], raw['is_active'],
                   raw['version'])

    def set_resource_types(self, resource_types):
        self.resource_types = resource_types

    def set_resource_model(self, resource_model):
        self.resource_model = resource_model

    def to_json(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=4)

    def save(self, db_connection):
        db_connection.save_service_template(self)


class Subscription:
    version = '0.1'

    def __init__(self, subscription_id, name, st_id,
                 is_active, resource_types, owner_id, resource_model, oaci_status, migration_status, migration_type):
        self.subscription_id = subscription_id
        self.name = name
        self.st_id = st_id
        self.is_active = is_active
        self.resource_types = resource_types
        self.owner_id = owner_id
        self.resource_model = resource_model
        self.oaci_status = oaci_status
        self.migration_status = migration_status
        self.migration_type = migration_type

    @classmethod
    def from_raw(cls, raw):
        return cls(raw['subscription_id'], raw['name'], raw['st_id'],
                   raw['is_active'], raw['resource_types'],
                   raw['owner_id'], 'CI', 'active', 'evaluated', 'regular')

    @classmethod
    def from_db(cls, db_connection, sub_id):
        subscription_id, customer_id, name, service_template_id, is_active, resource_model, oaci_status, migration_status, migration_type \
            = db_connection.get_subscription_from_db(sub_id=sub_id)[0]

        return cls(subscription_id=subscription_id, name=name,
                   st_id=service_template_id, is_active=is_active,
                   resource_types=[], owner_id=customer_id,
                   resource_model=resource_model,
                   oaci_status=oaci_status, migration_status=migration_status,
                   migration_type=migration_type)

    def set_resource_types(self, resource_types):
        self.resource_types = resource_types

    def set_resource_model(self, resource_model):
        self.resource_model = resource_model

    def set_migration_status(self, migration_status):
        self.migration_status = migration_status

    def set_migration_type(self, migration_type):
        self.migration_type = migration_type

    def to_json(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=4)

    def save(self, db_connection):
        db_connection.save_subscription(self)


class VirtualServerTemplate:
    version = '0.1'

    def __init__(self, name, description):
        self.aps = {
            "type": "http://www.parallels.com/infrastructure/virtual-environment/payg-sample/1.0",
            "revision": 0
        }
        self.name = name
        self.description = description
        self.osGroups = []
        self.defaultOsGroup = "Linux"

        self.appTemplates = []

        self.presets = []
        self.defaultPreset = "Web Server"
        self.ramTicks = [512, 1024, 4096, 8192]
        self.bandwidthTicks = [10, 100, 10000]
        self.ramLimit = {}
        self.diskLimit = {}
        self.cpuPowerLimit = {}
        self.cpuCoresLimit = {}

    def to_json(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=4)


class OsGroup:
    version = '0.1'

    def __init__(self, name, default_os):
        self.name = name
        self.defaultOs = default_os
        self.osList = []
        self.storageTypes = ['local', 'pcs']


class OsTemplate:
    version = '0.1'

    def __init__(self, name, priority):
        # example: centos-7-x86_64-pcs6
        self.name = name
        # example: 0, 1, 2
        self.priority = priority


class Preset:
    version = '0.1'

    def __init__(self, name, cpu_number, cpu_power, disk_space, ram, bandwidth):
        self.name = name
        self.cpuNumber = cpu_number
        self.cpuPower = cpu_power
        self.diskSpace = disk_space
        self.ram = ram
        self.bandwidth = bandwidth


class Limit:
    version = '0.1'

    def __init__(self, minimum, maximum, step):
        self.min = minimum
        self.max = maximum
        self.step = step


class ApsId:
    version = '0.1'

    def __init__(self):
        self.id = ''

    def set_id(self, id):
        self.id = id


class ApsTypeWithSubscription:
    version = '0.1'

    def __init__(self):
        self.type = "http://www.parallels.com/paci/vps-hosting/payg/1.2"
        self.subscription = ApsId()


class Usage:
    version = '0.1'

    def __init__(self):
        self.usage = 0

    def update_usage(self, value):
        self.usage += value


class HostingPayg:
    version = '0.1'

    def __init__(self):
        self.aps = ApsTypeWithSubscription()
        self.ramRunningPAYG = Usage()
        self.ramStoppedPAYG = Usage()
        self.hddRunningPAYG = Usage()
        self.hddStoppedPAYG = Usage()
        self.hddPcsRunningPAYG = Usage()
        self.hddPcsStoppedPAYG = Usage()
        self.cpuPowerContainerRunningPAYG = Usage()
        self.cpuPowerContainerStoppedPAYG = Usage()
        self.cpuPowerVmRunningPAYG = Usage()
        self.cpuPowerVmStoppedPAYG = Usage()
        self.bandwidthRunningPAYG = Usage()
        self.bandwidthStoppedPAYG = Usage()
        self.outboundTraffic = Usage()
        self.inboundTraffic = Usage()
        self.backupSpacePAYG = Usage()
        self.imageSpacePAYG = Usage()

    def set_aps_property(self, sub_uuid):
        self.aps.subscription.set_id(sub_uuid)

    def sum_usage_ram_running_payg(self, usage):
        self.ramRunningPAYG.update_usage(usage)

    def sum_usage_ram_stopped_payg(self, usage):
        self.ramStoppedPAYG.update_usage(usage)

    def sum_usage_hdd_running_payg(self, usage):
        self.hddRunningPAYG.update_usage(usage)

    def sum_usage_hdd_stopped_payg(self, usage):
        self.hddStoppedPAYG.update_usage(usage)

    def sum_usage_hdd_pcs_running_payg(self, usage):
        self.hddPcsRunningPAYG.update_usage(usage)

    def sum_usage_hdd_pcs_stopped_payg(self, usage):
        self.hddPcsStoppedPAYG.update_usage(usage)

    def sum_usage_cpu_power_container_running_payg(self, usage):
        self.cpuPowerContainerRunningPAYG.update_usage(usage)

    def sum_usage_cpu_power_container_stopped_payg(self, usage):
        self.cpuPowerContainerStoppedPAYG.update_usage(usage)

    def sum_usage_cpu_power_vm_running_payg(self, usage):
        self.cpuPowerVmRunningPAYG.update_usage(usage)

    def sum_usage_cpu_power_vm_stopped_payg(self, usage):
        self.cpuPowerVmStoppedPAYG.update_usage(usage)

    def sum_usage_bandwidth_running_payg(self, usage):
        self.bandwidthRunningPAYG.update_usage(usage)

    def sum_usage_bandwidth_stopped_payg(self, usage):
        self.bandwidthStoppedPAYG.update_usage(usage)

    def sum_usage_outbound_traffic(self, usage):
        self.outboundTraffic.update_usage(usage)

    def sum_usage_inbound_traffic(self, usage):
        self.inboundTraffic.update_usage(usage)

    def sum_usage_backup_space(self, usage):
        self.backupSpacePAYG.update_usage(usage)

    def sum_usage_image_space(self, usage):
        self.imageSpacePAYG.update_usage(usage)

    def delete_attribute_if_usage_zero(self, attribute):
        if hasattr(getattr(self, attribute), "usage") and getattr(self, attribute).usage == 0:
            delattr(self, attribute)

    def to_json(self):
        map(lambda attr: self.delete_attribute_if_usage_zero(attr), self.__dict__.keys())
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=4)
