#!/usr/bin/python
# -*- coding: utf-8 -*-

import argparse

from poaupdater import uLogging as log
import logging_utils

parser = argparse.ArgumentParser()
parser.add_argument("--status", action="store_true", help="get status of subscriptions and migration itself")
parser.add_argument("--subscriptions_file", type=str, help="get subscriptions from file")
parser.add_argument("--subscriptions", nargs='+', type=int, help="list of subscriptions to migrate")
# format is:
# python main.py --subscriptions 1000053 1000054 1000055 --stage finalize
parser.add_argument("--stage", type=str, help="perform stage of migration",
                    choices=['evaluation', 'injection', 'reinjection', 'provisioning', 'finalize', 'rollback', 'fast', 'only-ui'])
parser.add_argument("--oaci", type=str, help="choose the state of subscription to be migrated, "
                                             "if it has virtual environments or not,"
                                             "or had active usage in past configured period",
                    choices=['inactive', 'was_active', 'active'])
parser.add_argument("--batch", type=int, help="maximum amount of subscription to be migrated in current session, "
                                              "default is 100")
args = parser.parse_args()

command_argument = 'status' if args.status else args.stage

log_path = logging_utils.log_path(command_argument)
logging_utils.setup_logging(file=log_path)
log.info("Migration script started, see detailed logs: {}".format(log_path))

from utils import *
from provisioning_utils import provide_payg_service, rollback_usage
from paciapsapi import revert_aps_subscription, create_payg_sample
from billingapi import *
import sys


def evaluation(sub_ids, migration_type, batch_size=default_batch_size):
    service_templates, virtual_server_templates = parse_service_templates()
    for virtual_server_template in virtual_server_templates:
        if is_payg_sample_already_exist(virtual_server_template.name):
            continue
        else:
            create_payg_sample(virtual_server_template)

    log.debug("[evaluation] parsed service templates and virtual server templates")

    # check if subscriptions already evaluated and in db
    subscriptions = get_all_subs()

    n = 0
    for subscription in subscriptions:
        if subscription.subscription_id in sub_ids[n:batch_size+n] \
                and subscription.migration_status == 'evaluated':
            log.debug("[evaluation] subscription {} already evaluated".format(subscription.subscription_id))
            n += 1
            continue
        else:
            continue
    sub_ids = sub_ids[n:batch_size+n]

    if len(sub_ids) == 0:
        return 0

    subscriptions = parse_subscriptions(sub_ids)
    if len(subscriptions) == 0:
        log.debug("[evaluation] none of specified subscriptions could be evaluated")
        sys.exit(1)

    log.debug("[evaluation] parsed subscriptions {}".format([i.subscription_id for i in subscriptions]))

    set_and_save_subscriptions(subs=subscriptions, sts=service_templates, migration_type=migration_type)
    for subscription in subscriptions:
        try:
            check_if_oaci_activated(customer_id=subscription.owner_id, subscription_id=subscription.subscription_id)
        except Exception as e:
            log.err("[evaluation] failed to evaluate subscription {}, "
                    "please check connection to Instance Manager and it's health, {}".format(subscription.subscription_id, e))
            sys.exit(1)


def injection(sub_ids, oaci_status='active', batch_size=default_batch_size, retry=False):
    subscriptions = get_all_subs()
    if len(subscriptions) == 0:
        log.warn("[injection] none of specified subscriptions are in migration DB, please check DB or re-evaluate")
        sys.exit(0)

    could_be_migrated = []
    allowed_statuses = ['evaluated']
    if retry is True:
        allowed_statuses.append('injected')

    for subscription in subscriptions:
        if subscription.subscription_id in sub_ids \
                and subscription.migration_status in allowed_statuses \
                and subscription.oaci_status == oaci_status:
            could_be_migrated.append(subscription.subscription_id)

    if len(could_be_migrated) == 0:
        log.warn("[injection] none of specified subscriptions is in {} state(s) and could not be injected"
                 .format(allowed_statuses))
        sys.exit(0)

    log.debug("[injection] these of specified subscriptions are in {} state(s) and could be injected {}"
              .format(allowed_statuses, could_be_migrated))
    try:
        inject_resources_into_subscriptions(sub_ids=could_be_migrated[:batch_size])
    except Exception as e:
        log.err("[injection] failed to inject resources into subscriptions, "
                "please verify 'resources.py' and resources in Service Template, {}".format(e))
        sys.exit(1)
    log.debug("[injection] injected resources into specified subscriptions")


def provisioning(sub_ids, oaci_status='active', batch_size=default_batch_size):
    subscriptions = get_all_subs()
    if len(subscriptions) == 0:
        log.warn("[provisioning] none of specified subscriptions are in migration DB, please check DB or re-evaluate")
        sys.exit(0)

    could_be_migrated = []
    for subscription in subscriptions:
        if subscription.subscription_id in sub_ids \
                and subscription.migration_status == 'injected' \
                and subscription.oaci_status == oaci_status:
            could_be_migrated.append(subscription)

    if len(could_be_migrated) == 0:
        log.warn("[provisioning] none of specified subscriptions are in 'injected' state and could be provisioned")
        sys.exit(0)

    log.debug("[provisioning] these of specified subscriptions are in 'injected' state and could be provisioned {}".format([i.subscription_id for i in could_be_migrated]))
    try:
        provide_payg_service(could_be_migrated[:batch_size])
    except Exception as e:
        log.err(traceback.format_exc())
        log.err("[provisioning] failed to provide CS PAYG resources into subscriptions, "
                "please verify 'resources.py' and resources in Service Template, {}".format(e))
        sys.exit(1)
    log.debug("[provisioning] provided resources for specified subscriptions")


def finalize(sub_ids, oaci_status='active', batch_size=default_batch_size):
    subscriptions = get_all_subs()
    if len(subscriptions) == 0:
        log.warn("[finalize] none of specified subscriptions are in migration DB, please check DB or re-evaluate")
        sys.exit(0)

    could_be_migrated = []
    for subscription in subscriptions:
        if subscription.subscription_id in sub_ids \
                and subscription.migration_status == 'provisioned' \
                and subscription.oaci_status == oaci_status:
            could_be_migrated.append(subscription)

    if len(could_be_migrated) == 0:
        log.warn("[finalize] none of specified subscriptions is in 'provisioned' state and could not be finalized")
        sys.exit(0)

    log.debug("[finalize] these of specified subscriptions are in 'provisioned' state and could be finalized {}".format([i.subscription_id for i in could_be_migrated]))
    try:
        finalize_migration(could_be_migrated[:batch_size])
    except Exception as e:
        log.err("[finalize] failed to finish migration of subscriptions, "
                "please verify 'resources.py' and resources in Service Plan, {}".format(e))
        sys.exit(1)
    log.debug("[finalize] finished migration for subscriptions")


def fast_turnoff_ui(sub_ids, oaci_status='active', batch_size=default_batch_size):
    subscriptions = get_all_subs()
    if len(subscriptions) == 0:
        log.warn("[fast_turnoff_ui] none of specified subscriptions are in migration DB, please check DB or re-evaluate")
        sys.exit(0)

    could_be_migrated = []
    for subscription in subscriptions:
        if subscription.subscription_id in sub_ids \
                and subscription.migration_status == 'injected' \
                and subscription.oaci_status == oaci_status:
            could_be_migrated.append(subscription)

    if len(could_be_migrated) == 0:
        log.warn("[fast_turnoff_ui] none of specified subscriptions are in 'injected' state and could turn old ui off")
        sys.exit(0)

    log.debug("[fast_turnoff_ui] these of specified subscriptions are in 'injected' state and could turn old ui off {}".format([i.subscription_id for i in could_be_migrated]))
    try:
        turnoff_ui_only(could_be_migrated[:batch_size])
    except Exception as e:
        log.err("[fast_turnoff_ui] failed to finish migration of subscriptions, "
                "please verify 'resources.py' and resources in Service Plan, {}".format(e))
        sys.exit(1)
    log.debug("[fast_turnoff_ui] finished migration for subscriptions")


def rollback(sub_ids):
    subscriptions = get_all_subs()
    if len(subscriptions) == 0:
        log.warn("[rollback] none of specified subscriptions are in migration DB, please check DB or re-evaluate")
        sys.exit(0)

    could_be_reverted = []
    for subscription in subscriptions:
        if subscription.subscription_id in sub_ids \
                and subscription.migration_status == 'provisioned':
            could_be_reverted.append(subscription)

    if len(could_be_reverted) == 0:
        log.warn("[rollback] none of specified subscriptions are in 'provisioned' state and could be reverted back")
        sys.exit(0)

    log.debug("[rollback] these of specified subscriptions are in 'provisioned' state and could be reverted back {}"
              .format([i.subscription_id for i in could_be_reverted]))

    local_db_connection = MigrationDatabase()
    for subscription in could_be_reverted:
        revert_aps_subscription(subscription_id=subscription.subscription_id)
        rollback_usage(subscription_id=subscription.subscription_id, local_db_connection=local_db_connection)
        log.debug("[rollback] subscription {} has been rolled back to CI model".format(subscription.subscription_id))
        db_connection.update_subscription_status(sub_id=subscription.subscription_id, status='injected')


def status():
    get_status()


if args.subscriptions_file:
    filename = args.subscriptions_file
    log.debug("reading subscriptions file")

    if not os.path.exists(filename):
        log.err("Unable to find file [{0}]".format(filename))
        sys.exit(1)

    with open(filename) as f:
        data = f.read()
    try:
        data = json.loads(data)
        log.debug("Valid json file, processing")
        subscriptions = data['subscriptions']
    except ValueError:
        log.err("Invalid json format, Please correct it using migration guide")
        sys.exit(1)

else:
    subscriptions = args.subscriptions

if args.oaci:
    oaci_status = args.oaci
else:
    oaci_status = 'active'

if args.batch:
    batch_size = args.batch
else:
    batch_size = default_batch_size


log.debug("batch size of current migration is: {}, reducing list of subscriptions to that size".format(batch_size))
if command_argument == "status":
    log.debug("getting status of subscriptions and migration itself")
    status()

elif command_argument == 'fast':
    log.debug("performing fast migration for subscriptions {}".format(subscriptions))
    evaluation(sub_ids=subscriptions, migration_type='fast', batch_size=batch_size)
    injection(subscriptions, oaci_status, batch_size=batch_size)
    provisioning(subscriptions, oaci_status, batch_size=batch_size)
    finalize(subscriptions, oaci_status, batch_size=batch_size)

elif command_argument == 'only-ui':
    evaluation(sub_ids=subscriptions, migration_type='only-ui', batch_size=batch_size)
    log.debug("performing fast ui turn off for subscriptions {} "
              "please note that CS resource will be injected but not provisioned".format(subscriptions))
    injection(subscriptions, oaci_status, batch_size=batch_size)
    fast_turnoff_ui(subscriptions, oaci_status, batch_size=batch_size)


elif command_argument == 'evaluation':
    log.debug("collecting information about CI subscriptions {} and it's Service Template".
              format(subscriptions))
    evaluation(sub_ids=subscriptions, migration_type='regular', batch_size=batch_size)

elif command_argument == 'injection':
    log.debug("injecting CS resources into subscriptions {}".
              format(subscriptions))
    injection(subscriptions, oaci_status, batch_size=batch_size)

elif command_argument == 'reinjection':
    log.debug("reinjecting CS resources into subscriptions {}".
              format(subscriptions))
    injection(subscriptions, oaci_status, batch_size=batch_size, retry=True)

elif command_argument == 'provisioning':
    log.debug("provisioning CS APS resources for subscriptions {}".
              format(subscriptions))
    provisioning(subscriptions, oaci_status, batch_size=batch_size)


elif command_argument == 'finalize':
    log.debug("finalizing the migration, performing billing tuning"
              "and turning old usage reporting off for subscriptions {}".
              format(subscriptions))
    finalize(subscriptions, oaci_status, batch_size=batch_size)


elif command_argument == 'rollback':
    log.debug("rolling back migrated subscriptions {}".
              format(subscriptions))
    rollback(subscriptions)


log.info("Migration script ended, see detailed logs: {}".format(log_path))
