#!/usr/bin/python
# -*- coding: utf-8 -*-

import sqlite3
from sqlite3 import Error
import os
import traceback
from config import database
from poaupdater import uLogging

sql_create_resources_table = """ CREATE TABLE IF NOT EXISTS resources (
                                    resource_type_id integer PRIMARY KEY,
                                    name text NOT NULL,
                                    resclass_name text NOT NULL,
                                    measurement_type text NOT NULL
                                ); """

sql_create_activation_parameters_table = """ CREATE TABLE IF NOT EXISTS activation_parameters (
                                    activation_parameter_id integer PRIMARY KEY AUTOINCREMENT,
                                    resource_type_id integer NOT NULL,
                                    service_template_id integer NOT NULL,
                                    var_name text NOT NULL,
                                    var_value text NOT NULL,
                                    UNIQUE(resource_type_id, service_template_id, var_name)       
                                ); """

sql_create_service_templates_table = """ CREATE TABLE IF NOT EXISTS service_templates (
                                    service_template_id integer PRIMARY KEY,
                                    name text NOT NULL,
                                    resource_model text NOT NULL,
                                    active integer NOT NULL
                                ); """

sql_create_subscriptions_table = """ CREATE TABLE IF NOT EXISTS subscriptions (
                                    subscription_id integer PRIMARY KEY,                                    
                                    customer_id integer NOT NULL,
                                    name text NOT NULL,
                                    service_template_id integer NOT NULL,
                                    is_active integer NOT NULL,
                                    resource_model text NOT NULL,                                
                                    oaci_status text NOT NULL,
                                    migration_status text NOT NULL,
                                    migration_type text NOT NULL
                                ); """

sql_rollback_usage = """ CREATE TABLE IF NOT EXISTS rollback_usage (
                                    id integer PRIMARY KEY AUTOINCREMENT,
                                    subscription_id integer NOT NULL,                                    
                                    sql text NOT NULL UNIQUE
                                ); """


def log_query(sql, params):
    if params:
        uLogging.debug("Local DB statement: {}, params: {}".format(sql, params))
    else:
        uLogging.debug("Local DB statement: {}".format(sql))


def create_table(conn, create_table_sql):
    """ create a table from the create_table_sql statement
    :param conn: Connection object
    :param create_table_sql: a CREATE TABLE statement
    :return:
    """
    log_query(create_table_sql, None)
    try:
        c = conn.cursor()
        c.execute(create_table_sql)
    except Error as e:
        uLogging.err(traceback.format_exc())
        raise e


def create_connection(db_file):
    """ create a database connection to a SQLite database
    :param db_file: file name with path, where sqllite3 db will be
    :return:
    """
    try:
        conn = sqlite3.connect(db_file)
        return conn
    except Error as e:
        uLogging.err(traceback.format_exc())
    return None


def remove_db():
    """
      Removes migration database, on failures upon evaluation stage
      :return:
      """
    if os.path.exists(database):
        os.remove(database)


class MigrationDatabase:
    version = '0.1'

    def __init__(self):
        self.conn = None
        """
          Initializes migration database, creates tables or return connection if DB exists
          :return:
          """
        # create a database and connection to db
        if os.path.exists(database):
            self.conn = create_connection(database)
        self.conn = create_connection(database)

        create_table(self.conn, sql_create_subscriptions_table)
        create_table(self.conn, sql_create_service_templates_table)
        create_table(self.conn, sql_create_resources_table)
        create_table(self.conn, sql_create_activation_parameters_table)
        create_table(self.conn, sql_rollback_usage)

    def execute(self, sql, params):
        cur = self.conn.cursor()
        log_query(sql, params)
        cur.execute(sql, params)
        self.conn.commit()

    def get(self, sql, params):
        cur = self.conn.cursor()
        if params is None:
            log_query(sql, None)
            cur.execute(sql)
        else:
            log_query(sql, params)
            cur.execute(sql, params)
        return cur.fetchall()

    def save_resource_type(self, resource_type):
        """
        Inserts Resource Type in DB
        :param resource_type:
        :return:
        """
        sql = '''
        INSERT OR IGNORE INTO resources(resource_type_id, name, resclass_name, measurement_type) VALUES(?,?,?,?) 
        '''
        self.execute(sql, (
            resource_type.resource_type_id,
            resource_type.name,
            resource_type.resclass_name,
            resource_type.measurement_type)
                     )

    def save_activation_parameter(self, activation_parameter):
        """
        Inserts Activation Parameter in DB
        :param activation_parameter: ActivationParameter Object, created in evaluation stage
        :return:
        """
        sql = ''' INSERT OR IGNORE INTO activation_parameters(
        resource_type_id, service_template_id, var_name, var_value) VALUES(?,?,?,?) '''

        self.execute(sql, (activation_parameter.resource_type_id, activation_parameter.service_template_id,
                           activation_parameter.var_name, activation_parameter.var_value))

    def save_service_template(self, service_template):
        """
        Inserts Service Template in DB
        :param service_template: ServiceTemplate Object, created in evaluation stage
        :return:
        """
        sql = ''' INSERT OR IGNORE INTO service_templates(service_template_id, name, resource_model, active)
        VALUES(?,?,?,?) '''

        self.execute(sql, (service_template.st_id, service_template.name,
                           service_template.resource_model, service_template.active))

    def save_subscription(self, subscription):
        """
        Inserts Subscription in DB
        :param subscription: Subscription Object, created in evaluation stage
        :return:
        """
        sql = '''INSERT OR IGNORE INTO subscriptions(
        subscription_id, customer_id, name, service_template_id, is_active, resource_model, oaci_status, migration_status, migration_type) 
        VALUES(?,?,?,?,?,?,?,?,?)'''

        self.execute(sql, (subscription.subscription_id, subscription.owner_id, subscription.name, subscription.st_id,
                           subscription.is_active, subscription.resource_model, subscription.oaci_status,
                           subscription.migration_status, subscription.migration_type))

    def update_subscription_status(self, sub_id, status):
        """
         Updates Subscription migration status in DB
         :param sub_id: Subscription id
         :param status: status of migration
         :return:
         """
        sql = '''UPDATE subscriptions SET migration_status = ? WHERE subscription_id = ?'''
        self.execute(sql, (status, sub_id))

    def update_subscription_oaci_status(self, sub_id, status):
        """
          Updates Subscription OACI status in DB, (f.e. Subscription without VE is inactive)
          :param sub_id: Subscription id
          :param status: status of migration
          :return:
          """
        sql = '''UPDATE subscriptions SET oaci_status = ? WHERE subscription_id = ?'''
        self.execute(sql, (status, sub_id))

    def get_subscription_from_db(self, sub_id):
        """
          Gets subscription from migration Database and return plain row
          :param sub_id: Subscription id
          :return:
          """
        sql = '''SELECT * FROM subscriptions WHERE subscription_id = ?'''
        row = self.get(sql, (sub_id,))
        return row

    def get_all_subscription_from_db(self):
        """
          Gets all subscriptions from migration Database and return plain row
          :return:
          """
        sql = '''SELECT subscription_id FROM subscriptions'''
        rows = self.get(sql, None)
        return rows

    def save_rollback_usage_sql(self, subscription_id, sql):
        """
        Inserts sql to execute in case of rollback in DB
        :param subscription_id: Subscription Object, created in evaluation stage
        :param sql: sql request to execute in case of rollback
        :return:
        """
        db_sql = '''INSERT OR IGNORE INTO rollback_usage(subscription_id, sql) VALUES(?,?)'''
        self.execute(db_sql, (subscription_id, sql))

    def get_rollback_usage_sql(self, subscription_id):
        """
          Gets all sql requests for subscription_id to execute in case of rollback
          :return: array of sql requests
          """
        sql = '''SELECT sql FROM rollback_usage WHERE subscription_id = ?'''
        rows = self.get(sql, (subscription_id,))
        result = []
        for row in rows:
            result.append(row[0])
        return result
