#!/usr/bin/python
# -*- coding: utf-8 -*-

import base64
import json
import socket
import traceback

import urllib2
from poaupdater import uLogging as log
from paci_openapi import get_paci_property
import sys
import datetime
from config import days_of_inactivity


def get_paci_settings():
    """
    Gets PACI RESTApi settings, url, admin user, pass
    :return:
    """
    """ gets im_url, im_user, im_pass from by openapi from OSA MN
    :return:
    """
    log.debug("Call openapi.OpenAPI method to get PACI settings")
    try:
        paci = {
            "im_url": get_paci_property(property_name="c2u.im.server.uri")["value"],
            "im_user": get_paci_property(property_name="rest.login")["value"],
            "im_pass": get_paci_property(property_name="rest.password")["value"]
        }
    except:
        log.err(traceback.format_exc())
        log.err("Failed to get PACI properties")
        sys.exit(1)
    return paci


def request_to_im(request, username, password, method='GET', data=None):
    """ send any request to IM
    :param request: full url of request to IM
    :param username: admin or user login in RestAPI
    :param password: password for RestAPI user
    :return:
    """

    class MethodRequest(urllib2.Request):
        def __init__(self, *args, **kwargs):
            if 'method' in kwargs:
                self._method = kwargs['method']
                del kwargs['method']
            else:
                self._method = None
            return urllib2.Request.__init__(self, *args, **kwargs)

        def get_method(self, *args, **kwargs):
            if self._method is not None:
                return self._method
            return urllib2.Request.get_method(self, *args, **kwargs)


    retries = 10

    http_handler = urllib2.HTTPHandler()
    opener = urllib2.build_opener(http_handler)

    # perform basic auth
    base64string = base64.encodestring('%s:%s' % (username, password))[:-1]

    # request
    json_data = None
    if data:
        json_data = json.dumps(data, sort_keys=False, indent=4)

    req = MethodRequest(request, data=json_data, method=method)
    req.add_header("Authorization", "Basic %s" % base64string)

    if data:
        req.add_header("Content-Type", "application/json")

    # get output in json format
    req.add_header("Accept", "application/json")

    while retries:
        try:
            url_handle = opener.open(req)
            response = json.loads(url_handle.read())
        except:
            retries -= 1
            log.err("Attempt to call IM number {}, failed".format(retries))
            continue
        log.debug("Called IM successfully. Retries left: {}".format(retries))
        return response
    log.err("No more attempts to call IM, request failed, please check IM connection and stability")
    return None


def get_customer_ves(paci, customer_id, subscription_id):
    """ Gets VEs of particular customer and subscription id
    :param paci: live connection to Instance Manager
    :param customer_id: customer ID
    :param subscription_id: subscription ID
    :return:
    """
    request = '{0}/of/{1}/ve/?subscription={2}'.format(paci["im_url"], customer_id, subscription_id)
    response = request_to_im(request, paci["im_user"], paci["im_pass"])
    return response


def get_os_templates(paci):
    """ Gets list of all OS template, registered in Instance Manager
    :param paci: live connection to Instance Manager
    :return:
    """
    request = '{0}/template'.format(paci["im_url"])
    response = request_to_im(request, paci["im_user"], paci["im_pass"])
    return response


def format_datetime_for_url(dt, tz):
    """
    Formats time to URL IM-request time with timezone
    :param dt: datetime
    :param tz: timezone from MN
    :return: formatted time for URL IM-request
    """
    return dt.strftime("%Y-%m-%d %H:%M ").replace(" ", '%20') + tz


def get_usage(paci, customer_id, subscription_id, time_from, time_to, timezone='+0000'):
    """
    Gets resource usage of particular customer and subscription for the time interval
    :param paci: live connection to Instance Manager
    :param customer_id: customer ID
    :param subscription_id: subscription ID
    :param time_from: from inclusive datetime
    :param time_to: to exclusive datetime
    :param timezone: timezone from MN
    :return: IM ResourceUsageReport for current subscription and given period
    """
    time_from_str = format_datetime_for_url(time_from, timezone)
    time_to_str = format_datetime_for_url(time_to, timezone)

    request = '{0}/of/{1}/usage/{2}/{3}/{4}?include-detached-backups=true'.\
        format(paci["im_url"], customer_id, subscription_id, time_from_str, time_to_str)
    response = request_to_im(request, paci["im_user"], paci["im_pass"])
    return response


def get_usage_last_3_month(paci, customer_id, subscription_id):
    """
    Gets VEs usage of particular customer and subscription id for last 3 months
    :param paci: live connection to Instance Manager
    :param customer_id: customer ID
    :param subscription_id: subscription ID
    :return: IM ResourceUsageReport for current subscription for last 3 month
    """
    # this is the 3 months start date from now
    time_from = datetime.datetime.now() + datetime.timedelta(-days_of_inactivity)
    # current time + whole current day
    time_to = datetime.datetime.now().date() + datetime.timedelta(days=1)

    return get_usage(paci, customer_id, subscription_id, time_from, time_to)


def get_all_ci_usage(paci, customer_id, subscription_id, last_update_time, timezone):
    """
    Gets IM resource usage from beginning of time to last time of update subscription CI resources usage
    :param paci: live connection to Instance Manager
    :param customer_id: customer ID
    :param subscription_id: subscription ID
    :param last_update_time: last update time of CI resources for current subscription, that get from oss DB
    :param timezone: timezone from MN
    :return: IM ResourceUsageReport for current subscription
    """
    time_from = datetime.datetime(1970, 1, 1)
    return get_usage(paci, customer_id, subscription_id, time_from, last_update_time, timezone)

def get_subscription(paci, customer_id, subscription_id):
    request = '{0}/of/{1}/subscription/{2}'.format(paci["im_url"], customer_id, subscription_id)
    return request_to_im(request, paci["im_user"], paci["im_pass"])

def change_subscription(paci, customer_id, subscription_id, subscription_data):
    request = '{0}/of/{1}/subscription/{2}'.format(paci["im_url"], customer_id, subscription_id)
    return request_to_im(request, paci["im_user"], paci["im_pass"], method='PUT', data=subscription_data)