#!/usr/bin/python
# -*- coding: utf-8 -*-

import traceback

from poaupdater import apsapi
from paci_openapi import *
from poaupdater import uLogging

uLogging.log_to_console = True

mn = 'localhost'
SUBSCRIPTION_SRV = 'http://parallels.com/aps/types/pa/subscription/1.0'
PAYG_SRV = 'http://www.parallels.com/paci/vps-hosting/1.0'
HEADERS = {'Content-Type': 'application/json'}
MN_CERT = '/usr/local/pem/APS/certificates/poa.pem'


def get_aps_api():
    """
    Get Aps API for requests
    :return: Aps API for requests
    """
    return apsapi.API('https://{0}:6308/aps/2'.format(mn))


def get_subscription_uuid(subscription_id):
    """
    Get APS subscription uuid by subscription Id
    :param subscription_id: Subscription id
    :return: subscription uuid
    """
    return get_aps_api().GET(
        '/resources?implementing({0}),eq(subscriptionId,{1})'.format(SUBSCRIPTION_SRV, subscription_id),
        HEADERS, '', MN_CERT)[0]['aps']['id']


def provide_payg(customer_id, subscription_id, payg_hosting):
    """
    Provides payg resource on APS bus for particular Subscription and Customer
    It triggers provisioning
    :param customer_id: Customer id
    :param subscription_id: Subscription id
    :param payg_hosting: Json of resource payg hosting with usage for counters
    :return:
    """
    token = get_aps_token(customer_id, subscription_id)['aps_token']
    headers = {'Content-Type': 'application/json', 'APS-Token': token}
    get_aps_api().POST('/resources', headers, payg_hosting)
    return 0


def attach_im_subscription(subscription_id):
    """
    Performs attaching of legacy CI subscription into CS APS subscription
    It triggers provisioning
    :param subscription_id: Subscription id
    :return:
    """
    aps_api = get_aps_api()

    payg_uid = get_payg_uid(subscription_id)

    detach = '/resources/{}/detach_im_subscription'.format(payg_uid)
    attach = '/resources/{}/attach_im_subscription/{}'.format(payg_uid, subscription_id)

    aps_api.POST(detach, HEADERS, '', MN_CERT)
    aps_api.POST(attach, HEADERS, '', MN_CERT)
    return 0


def get_payg_uid(subscription_id):
    """
    Get APS resource payg hosting uuid by subscription Id
    :param subscription_id: Subscription id
    :return: payg hosting uuid
    """
    payg_uid = get_aps_api().GET('/resources?implementing({0}),eq(subId,{1})'.
                                 format(PAYG_SRV, subscription_id), HEADERS, '', MN_CERT)[0]['aps']['id']
    return payg_uid


def revert_aps_subscription(subscription_id):
    """
    Performs detaching of CS APS subscription and deletes payg resource
    Used for rollback during migration
    :param subscription_id: Subscription id
    :return:
    """
    try:
        aps_api = get_aps_api()
        payg_uid = get_payg_uid(subscription_id)

        detach = '/resources/{}/detach_im_subscription'.format(payg_uid)
        delete = '/resources/{}'.format(payg_uid)

        aps_api.POST(detach, HEADERS, '', MN_CERT)
        aps_api.DELETE(delete, HEADERS, '', MN_CERT)
    except Exception as e:
        uLogging.debug("failed to revert APS subscription, {}\n{}".format(e, traceback.format_exc()))
    return 0


def create_payg_sample(virtual_server_template):
    """
    Creates payg-sample resource based on VirtualServerTemplate class
    :param virtual_server_template: VST, generated on evaluation stage
    :return:
    """
    aps_api = get_aps_api()
    paci_cert = '/usr/local/share/PACI-aps/paci.pem'

    payg_sample = virtual_server_template.to_json()
    r = aps_api.POST('/application/samples-payg', HEADERS, payg_sample, paci_cert)
    vst_aps_id = r['aps']['id']

    created = '/resources/{}/created'.format(vst_aps_id)
    aps_api.PUT(created, HEADERS, '', paci_cert)
    return 0
