import threading
import sys
import traceback

import paciapi
import config

from utils import get_subscriptions_chunks
from mappers import MigrationDatabase
from classes import HostingPayg
from oss_utils import turnoff_old_usage_reporting, drop_usage_for_service_resources
from paciapsapi import provide_payg, attach_im_subscription, revert_aps_subscription, get_subscription_uuid
from paciapi import get_all_ci_usage, get_paci_settings
from paci_openapi import get_subscription
from resources import get_drop_usage_list, DROP_USAGE
from poaupdater import uSysDB
from poaupdater import uLogging as log
from collections import OrderedDict

paci = get_paci_settings()

default_drop_usage_list = get_drop_usage_list()
if len(DROP_USAGE) != len(default_drop_usage_list):
    log.err("please compare resources.py:DROP_USAGE with Resources names on your platform")
    sys.exit(1)


def provide_payg_service(sub_ids):
    """
    Provide PAYG service and attach CI subscription to APS for specified Subscriptions
    :param sub_ids: List of Subscriptions ids
    :return:
    """
    subs = []

    aps_api_threads = get_config_parameter('aps_api_threads')
    if len(sub_ids) <= aps_api_threads:
        threads_num = 1
    else:
        threads_num = aps_api_threads

    chunks = get_subscriptions_chunks(sub_ids, len(sub_ids) / threads_num)
    threads = []
    results = {}
    for i in range(threads_num):
        t = threading.Thread(target=provide_payg_worker, args=[chunks[i], results, i])
        threads.append(t)
        t.start()

    for t in threads:
        t.join()

    for result in results:
        subs.extend(results[result])

    db_connection = MigrationDatabase()
    for sub_id in subs:
        db_connection.update_subscription_status(sub_id=sub_id, status='provisioned')


def provide_payg_worker(chunk, results, thread_id):
    """
    Turn off usage reporting of CI resources by subscription,
    provide PAYG service with usage for counter, attach CI IM subscription to APS
    and drop usage for application service resources for specified Subscriptions in Thread
    :param chunk: List of chunk of Subscriptions ids
    :param results: Dict, where each thread will store results
    :param thread_id: id of the Thread
    :return:
    """
    subs = []
    oa_con = uSysDB.connect()
    # new instance for each thread
    local_db_connection = MigrationDatabase()

    for subscription in chunk:
        sub_id = subscription.subscription_id
        provide_precheck(sub_id)
        try:
            turnoff_old_usage_reporting(subscription_id=sub_id, con=oa_con, local_db_connection=local_db_connection)
            log.debug("[provisioning] all old usage reporting has been turned off for subscription {} ".format(sub_id))

            payg_hosting = get_payg_hosting_with_usage_for_counters(con=oa_con, subscription_id=sub_id,
                                                                    customer_id=subscription.owner_id)
            log.debug("[provisioning] resource payg hosting ready for create: {}".format(payg_hosting))
            provide_payg(customer_id=subscription.owner_id, subscription_id=sub_id, payg_hosting=payg_hosting)
            log.debug("[provisioning] payg resource has been provisioned for subscription {}".format(sub_id))

            attach_im_subscription(subscription_id=sub_id)
            log.debug("[provisioning] legacy IM subscription {} has been attached to CS model".format(sub_id))

            modify_im_subscription(customer_id=subscription.owner_id, subscription_id=sub_id)

            resource_type_ids = []
            for entity in default_drop_usage_list:
                resource_type_ids.append(entity['resource_id'])
            drop_usage_for_service_resources(con=oa_con, subscription_id=sub_id, resource_type_ids=resource_type_ids,
                                             local_db_connection=local_db_connection)
            log.debug("[provisioning] CI resource usage for resources {} was dropped for subscription {}".format(
                resource_type_ids, sub_id))

            subs.append(sub_id)
        except Exception as e:
            log.err(traceback.format_exc())
            log.err("[provisioning] failed one of operations for subscription {}, "
                    "please check Subscription, Service Plan, health of PACI APS Endpoint and Instance Manager "
                    "and rerun the stage, {}".format(subscription.subscription_id, e))
            rollback_usage(subscription_id=subscription.subscription_id, local_db_connection=local_db_connection)
            revert_aps_subscription(subscription_id=subscription.subscription_id)

    if len(subs) > 0:
        results[thread_id] = subs


def provide_precheck(subscription_id):
    # check if all resources to drop usage exist in the subscription
    resource_list = get_subscription(subscription_id)['resource_types']
    for resource_to_drop in default_drop_usage_list:
        resource_to_find_list = list(
            filter(lambda resource_to_find: resource_to_drop['resource_id'] == resource_to_find['resource_type_id'],
                   resource_list))
        if len(resource_to_find_list) == 0:
            raise Exception("Failed to find resource id: {} name: {} in the subscription {}."
                            .format(resource_to_drop['resource_id'],
                                    resource_to_drop['name'],
                                    subscription_id))


def get_payg_hosting_with_usage_for_counters(con, subscription_id, customer_id):
    """
    Collect and update resource usage from IM to APS subscription
    :param con: live connection to OA Sys DB
    :param subscription_id: Subscription id
    :param customer_id: Customer id
    :return:
    """
    # get last update time of CI resources for current subscription
    cur = con.cursor()
    cur.execute("SELECT last_update FROM subs_resources "
                "LEFT JOIN resource_types ON resource_types.rt_id = subs_resources.rt_id "
                "WHERE sub_id = {} AND restype_name LIKE 'CI %' "
                "ORDER BY last_update DESC "
                "LIMIT 1".format(subscription_id))
    last_update_time = cur.fetchone()[0]

    im_resource_usage = get_all_ci_usage(paci=paci, customer_id=customer_id, subscription_id=subscription_id,
                                         last_update_time=last_update_time, timezone=get_oa_timezone(con))
    subscription_uuid = get_subscription_uuid(subscription_id)

    return parse_hosting_payg(subscription_uuid, im_resource_usage).to_json()


def rollback_usage(subscription_id, local_db_connection, oa_con=uSysDB.connect()):
    """
    rollbacks operation with usage from oss DB
    :param subscription_id: Subscription id
    :param local_db_connection: local SQL connection
    :param oa_con: live connection to OA Sys DB
    :return:
    """
    try:
        sql_list = local_db_connection.get_rollback_usage_sql(subscription_id=subscription_id)
        cur = oa_con.cursor()
        for sql in sql_list:
            cur.execute(sql)
        oa_con.commit()
    except Exception as e:
        log.debug('failed to rollbacks operation with usage from oss DB, {}'.format(e))


def parse_hosting_payg(subscription_uuid, im_resource_usage):
    """
    Get resource usage from IM and set usage for APS resource Hosting PAYG
    :param subscription_uuid: APS subscription uuid
    :param im_resource_usage: ResourceUsageReport from IM for current subscription
    :return: resource usage for APS Hosting PAYG
    """
    hosting_payg = HostingPayg()
    hosting_payg.set_aps_property(sub_uuid=subscription_uuid)
    for ve_resource_usage in im_resource_usage.get('ve-resource-usage-report', []):
        resource_usage_list = ve_resource_usage.get('resource-usage', {})
        hosting_payg.sum_usage_ram_running_payg(
            get_usage_by_resource(resource_usage_list, 'ram-hours', 'while-running'))
        hosting_payg.sum_usage_ram_stopped_payg(
            get_usage_by_resource(resource_usage_list, 'ram-hours', 'while-stopped'))
        hosting_payg.sum_usage_hdd_running_payg(
            get_usage_by_resource(resource_usage_list, 'hdd-hours', 'while-running'))
        hosting_payg.sum_usage_hdd_stopped_payg(
            get_usage_by_resource(resource_usage_list, 'hdd-hours', 'while-stopped'))
        hosting_payg.sum_usage_hdd_pcs_running_payg(
            get_usage_by_resource(resource_usage_list, 'pcs-hours', 'while-running'))
        hosting_payg.sum_usage_hdd_pcs_stopped_payg(
            get_usage_by_resource(resource_usage_list, 'pcs-hours', 'while-stopped'))
        hosting_payg.sum_usage_bandwidth_running_payg(
            get_usage_by_resource(resource_usage_list, 'bandwidth-hours', 'while-running'))
        hosting_payg.sum_usage_bandwidth_stopped_payg(
            get_usage_by_resource(resource_usage_list, 'bandwidth-hours', 'while-stopped'))

        if '@technology' in ve_resource_usage and ve_resource_usage['@technology'] == 'VM':
            hosting_payg.sum_usage_cpu_power_vm_running_payg(
                get_usage_by_resource(resource_usage_list, 'cpu-hours', 'while-running'))
            hosting_payg.sum_usage_cpu_power_vm_stopped_payg(
                get_usage_by_resource(resource_usage_list, 'cpu-hours', 'while-stopped'))
        else:
            hosting_payg.sum_usage_cpu_power_container_running_payg(
                get_usage_by_resource(resource_usage_list, 'cpu-hours', 'while-running'))
            hosting_payg.sum_usage_cpu_power_container_stopped_payg(
                get_usage_by_resource(resource_usage_list, 'cpu-hours', 'while-stopped'))

        ve_traffic_list = ve_resource_usage.get('ve-traffic', {})
        hosting_payg.sum_usage_inbound_traffic(get_usage_by_traffic(ve_traffic_list, 'public-incoming'))
        hosting_payg.sum_usage_outbound_traffic(get_usage_by_traffic(ve_traffic_list, 'public-outgoing'))
        hosting_payg.sum_usage_backup_space(to_gb(im_resource_usage.get('backup-size-hours')))
        hosting_payg.sum_usage_image_space(im_resource_usage.get('images-size-hours'))

    return hosting_payg


def get_usage_by_resource(resource_usage_list, resource_type, resource_usage_type):
    """
    Find corresponding resource from IM ve-resource-usage-report and return APS usage
    :param resource_usage_list: list resource-usage for VE
    :param resource_type: type of resource by hours
    :param resource_usage_type: one of resource type: 'while-running' or 'while-stopped'
    :return: usage
    """
    resource_usages = list(filter(lambda resource_usage:
                                  resource_usage['@resource-type'] == resource_type and
                                  resource_usage['@resource-usage-type'] == resource_usage_type,
                                  resource_usage_list))
    if len(resource_usages) > 0 and '@value' in resource_usages[0]:
        return resource_usages[0]['@value']
    else:
        return 0


def get_usage_by_traffic(ve_traffic_list, traffic_type):
    """
    Find corresponding resource traffic from IM ve-traffic and return APS usage
    :param ve_traffic_list: list ve-traffic for VE
    :param traffic_type: one of traffic type: 'incoming' or 'outgoing'
    :return: usage
    """
    ve_traffics = list(filter(lambda ve_traffic: ve_traffic['@traffic-type'] == traffic_type, ve_traffic_list))
    if len(ve_traffics) > 0 and '@used' in ve_traffics[0]:
        return ve_traffics[0]['@used']
    else:
        return 0


def get_oa_timezone(con):
    """
    Get timezone from Management Node for current time
    :param con: live connection to OA Sys DB
    :return: timezone from Management Node with format +HHMM
    """
    cur = con.cursor()
    cur.execute("select clock_timestamp()")
    timestamp_with_time_zone = cur.fetchone()
    return timestamp_with_time_zone[0].strftime("%z")


def modify_im_subscription(customer_id, subscription_id):
    if get_config_parameter('allow_all_storage_types_in_subscription', False):
        add_all_storage_types_in_subscription(customer_id, subscription_id)


def add_all_storage_types_in_subscription(customer_id, subscription_id):
    subscription = paciapi.get_subscription(paci, customer_id, subscription_id)

    storage_array = []
    storage_array.append({'@type': 'local', '@technology': 'CT', '@default': 'true'})
    storage_array.append({'@type': 'local', '@technology': 'VM', '@default': 'true'})
    storage_array.append({'@type': 'pcs', '@technology': 'CT'})
    storage_array.append({'@type': 'pcs', '@technology': 'VM'})

    # properties must be in same order as in c2u.xsd in IM
    subscription_dto = OrderedDict([
    (get_property_as_key_value_pair(subscription, '@id')),
    (get_property_as_key_value_pair(subscription, 've-limit')),
    (get_property_as_key_value_pair(subscription, 'lb-limit')),
    (get_property_as_key_value_pair(subscription, 'technology')),
    (get_property_as_key_value_pair(subscription, 'public-ip-limit')),
    (get_property_as_key_value_pair(subscription, 'public-ipv6-limit')),
    ('storage', storage_array),
    (get_property_as_key_value_pair(subscription, 'backup-schedule-info')),
    (get_property_as_key_value_pair(subscription, 'policy')),
    (get_property_as_key_value_pair(subscription, 'provisioning-attribute'))])

    paciapi.change_subscription(paci, customer_id, subscription_id, subscription_dto)


def get_property_as_key_value_pair(object, property_name):
    value = None
    try:
        value = object[property_name]
    except:
        pass
    return (property_name, value)

def get_config_parameter(name, default_value=None):
    if hasattr(config, name):
        return getattr(config, name)

    else:
        return default_value

def to_gb(bytes):
    if bytes is None:
        return 0
    else:
        return int(bytes / (1024 * 1024 * 1024))