#!/usr/bin/python
# -*- coding: utf-8 -*-
from classes import *
from resources import *
from paciapi import *
from billingapi import *
from paci_openapi import *
import sys
import threading
import re
import os
from shutil import copyfile
from config import database
from mappers import MigrationDatabase

VERSION_TXT = 'version.txt'
DB_BACKUP_TEMPLATE = '/var/log/pa/ci2cs/pypaci_backup_{}.db'

try:
    from poaupdater import openapi
    from poaupdater import uSysDB
    from poaupdater import uLogging as log
except:
    log.err("Please call the script from POA node")
    sys.exit(1)

try:
    pba_api = init_pba()
except:
    log.err("Failed to initiate PBA API")
    sys.exit(1)

try:
    paci = get_paci_settings()
except:
    log.err(traceback.format_exc())
    log.err("Failed to get PACI settings")
    sys.exit(1)

if os.path.exists("os_template_cache.json"):
    log.debug("There is existing os template cache, loading from json file...")
    with open("os_template_cache.json") as f:
        data = f.read()
    try:
        os_templates_cache = json.loads(data)
    except ValueError:
        log.err("Invalid json format, Please  fill with correct json containing os templates info")
        sys.exit(1)

else:
    log.debug("No existing os template cache, making new empty...")
    os_templates_cache = {}

default_cs_resources_tuple = get_default_cs_resources()

if len(cs_resources) != len(default_cs_resources_tuple):
    log.err("please compare resources.py:cs_resources with Resources names on your platform")
    sys.exit(1)


def backup_db_file():
    if os.path.exists(database):
        with open(VERSION_TXT, 'r') as version_file:
            version = version_file.read().strip()

        log.debug('script version: {}'.format(version))
        pattern = re.compile("^\d+\.\d+$")
        if pattern.match(version):
            copyfile(database, DB_BACKUP_TEMPLATE.format(version))
        else:
            raise Exception("Failed to parse {}, please check version format.".format(VERSION_TXT))


backup_db_file()
db_connection = MigrationDatabase()


def default_ve_size_restictions():
    cpu_limit = Limit(minimum=1000, maximum=3000, step=1000)
    ram_limit = Limit(minimum=512, maximum=8196, step=512)
    disk_limit = Limit(minimum=10, maximum=100, step=1)
    cpu_cores_limit = Limit(minimum=1, maximum=16, step=1)

    return cpu_limit, ram_limit, disk_limit, cpu_cores_limit


def parse_ve_size_restrictions(ve_size_restrictions):
    """
    Parses activation parameter of 've_size_restrictions' and makes Limit classes for Virtual Server Template
    :param ve_size_restrictions: json-able string from DB or openAPI output
    :return:
    """
    log.debug("[evaluation] trying to load Legacy OACI CI resources activation parameters json {},"
              "in order to create possible CS Sample Resource".format(format(ve_size_restrictions)))
    try:
        ve_size_restrictions = json.loads(ve_size_restrictions)

        tech = 'CT'
        if tech not in ve_size_restrictions.keys():
            tech = 'VM'

        space = 'local_space'
        if space not in ve_size_restrictions.keys():
            space = 'pcs_space'

        cpu_limit = Limit(minimum=ve_size_restrictions[tech]['min']['cpu_frequency'],
                          maximum=ve_size_restrictions[tech]['max']['cpu_frequency'],
                          step=ve_size_restrictions[tech]['step']['cpu_frequency'])

        ram_limit = Limit(minimum=ve_size_restrictions[tech]['min']['ram'],
                          maximum=ve_size_restrictions[tech]['max']['ram'],
                          step=ve_size_restrictions[tech]['step']['ram'])

        disk_limit = Limit(minimum=ve_size_restrictions[tech]['min'][space],
                           maximum=ve_size_restrictions[tech]['max'][space],
                           step=ve_size_restrictions[tech]['step'][space])

        cpu_cores_limit = Limit(minimum=ve_size_restrictions[tech]['min']['cpu_count'],
                                maximum=ve_size_restrictions[tech]['max']['cpu_count'],
                                step=ve_size_restrictions[tech]['step']['cpu_count'])
    except ValueError:
        log.warn("[evaluation] non critical error occurred on attempt to parse "
                 "Legacy OACI CI resources activation parameters : {}, "
                 "falling to defaults".format(ve_size_restrictions))
        cpu_limit, ram_limit, disk_limit, cpu_cores_limit = default_ve_size_restictions()

    log.debug("[evaluation] passed, json loaded")

    return cpu_limit, ram_limit, disk_limit, cpu_cores_limit


def parse_default_configuration(default_configurations):
    """
    Parses activation parameter of 'default_configurations' and makes Presets for Virtual Server Template
    (f.e. Web Server with cpu_count and etc)
    :param default_configurations: json-able string from DB or openAPI output
    :return:
    """
    presets = []
    default_configurations = json.loads(default_configurations)
    for default_configuration in default_configurations:
        presets.append(
            Preset(name=default_configuration['name'],
                   cpu_number=default_configuration['c2u.cpu-hours..cpu_count'],
                   cpu_power=default_configuration['c2u.cpu-hours..cpu_frequency'],
                   disk_space=default_configuration['c2u.storage-hours.initial_space'],
                   ram=default_configuration['c2u.RAM-hours..ram'], bandwidth=10000)
        )

    return presets


def is_payg_sample_already_exist(vst_name):
    """
    Call OA via OpenAPI for resources of service link and checks if name already exist
    :param vst_name: String name of Virtual Server Template
    :return:
    """
    rs = get_resource_by_class(resource_class_name='rc.saas.service.link', show_system_rt=False)
    result = False
    for r in rs:
        if r['resource_type_name'] == vst_name:
            return True
        else:
            continue
    return result


def generate_payg_sample(default_configurations, ve_size_restrictions, origin_service_template_id):
    """
    Calls parsers for default_configurations and ve_size_restrictions
    and return Virtual Server Template class (json-able)
    :param default_configurations: json-able string from DB or openAPI output
    :param ve_size_restrictions: json-able string from DB or openAPI output
    :param origin_service_template_id: id of Service Template from which VST will been created
    :return:
    """
    if not bool(os_templates_cache):
        log.debug("Os template cache is empty, calling IM to get os templates...")
        templates = get_os_templates(paci=paci)
        with open('os_template_cache.json', 'w') as outfile:
            log.debug("Saving os templates cache on filesystem...")
            json.dump(templates, outfile)
    else:
        templates = os_templates_cache

    os_templates = []
    for template in templates['template']:
        os_templates.append(OsTemplate(template['@name'], 0))

    os_group = OsGroup("Linux", "centos-7-x86_64-pcs6")

    os_group.osList.extend(os_templates)
    vst = VirtualServerTemplate("Virtual Server Template For {} Service Template".format(origin_service_template_id),
                                "Generated For Migration For {} Service Template".format(origin_service_template_id))
    vst.osGroups.append(os_group)

    presets = parse_default_configuration(default_configurations)
    vst.presets.extend(presets)
    vst.defaultPreset = presets[0].name

    log.debug(
        "[evaluation] going to parse_ve_size_restrictions of service template: {}".format(origin_service_template_id))
    cpu_limit, ram_limit, disk_limit, cpu_cores_limit = parse_ve_size_restrictions(ve_size_restrictions)

    vst.ramLimit = ram_limit
    vst.diskLimit = disk_limit
    vst.cpuPowerLimit = cpu_cores_limit
    vst.cpuPowerLimit = cpu_limit
    vst.cpuCoresLimit = cpu_cores_limit

    return vst


def parse_service_templates():
    """
    Takes all Service Templates from Odin Automation, parses activation parameters of included resources
    and save it in migration DB
    :return:
    """
    service_templates = get_all_service_templates()
    sts = []

    for st in service_templates:
        raw = get_service_template(st['st_id'])
        s = ServiceTemplate.from_raw(raw)

        rts = []
        for r in raw['resource_types']:
            res = ResourceType.from_raw(r)

            acts = []
            for act in r['act_params']:
                ap = ActivationParameter.from_raw(act)
                ap.set_resource_type_id(res.resource_type_id)
                ap.set_service_template_id(s.st_id)

                ap.save(db_connection)

                acts.append(ap)

            res.set_act_params(acts)
            res.set_attributes(r['attributes'])

            rts.append(res)

        s.set_resource_types(rts)
        sts.append(s)

    legacy_configurations = []

    for st in sts:
        resource_model_found = False
        legacy_configuration = {}
        for st.resource_type in st.resource_types:
            for st.resource_type.act_param in st.resource_type.act_params:
                if st.resource_type.act_param.var_name == 'service_id' \
                        and st.resource_type.act_param.var_value == 'vps-hosting':
                    st.resource_model = 'VPS'
                    resource_model_found = True
                    break
                elif st.resource_type.act_param.var_name == 've_size_restrictions':
                    legacy_configuration['ve_size_restrictions'] = st.resource_type.act_param.var_value
                    legacy_configuration['origin_service_template_id'] = st.st_id
                    st.resource_model = 'CI'
                    resource_model_found = True
                elif st.resource_type.act_param.var_name == 'default_configuration':
                    legacy_configuration['default_configuration'] = st.resource_type.act_param.var_value
                    legacy_configuration['origin_service_template_id'] = st.st_id

                    st.resource_model = 'CI'
                    resource_model_found = True
                    break
                elif st.resource_type.act_param.var_name == 'service_id' \
                        and st.resource_type.act_param.var_value == 'payg':
                    st.resource_model = 'CS'
                    resource_model_found = True
                    break
                elif st.resource_type.act_param.var_name == 'host.vps.cfg_id':
                    st.resource_model = 'Legacy'
                    resource_model_found = True
                    break
        if not resource_model_found:
            st.resource_model = 'Other'

        legacy_configurations.append(legacy_configuration)
        st.save(db_connection)

    vsts = []
    for legacy_configuration in legacy_configurations:
        if bool(legacy_configuration):
            vst = generate_payg_sample(default_configurations=legacy_configuration['default_configuration'],
                                       ve_size_restrictions=legacy_configuration['ve_size_restrictions'],
                                       origin_service_template_id=legacy_configuration['origin_service_template_id'])
            vsts.append(vst)

    return sts, vsts


def get_subscriptions_chunks(subscriptions_list, batch):
    """Yield successive n-sized chunks from l."""
    final_list_of_chunks = []
    for i in range(0, len(subscriptions_list), batch):
        final_list_of_chunks.append(subscriptions_list[i:i + batch])
    return final_list_of_chunks


def get_subscriptions_worker(chunk, results, thread_id):
    """
    :param chunk: Sliced list of subscriptions to evaluate
    :param results: Dict, where each thread will store results
    :param thread_id: id of the Thread
    :return:
    """
    subs = []

    for sub in chunk:
        try:
            raw = get_subscription(sub)
        except Exception as e:
            log.err(traceback.format_exc())
            log.err("[evaluation] cannot evaluate subscription {}, most likely it doesn't exist, {}".format(sub, e))
            continue
        s = Subscription.from_raw(raw)

        rts = []
        for r in raw['resource_types']:
            res = ResourceType.from_raw(r)
            rts.append(res)

        s.set_resource_types(rts)

        subs.append(s)
    if len(subs) > 0:
        results[thread_id] = subs


def parse_subscriptions(sub_ids):
    """
    Gets Subscriptions from Open API and parses Resources and it's parameters
    and save it in migration DB
    :param sub_ids: list of Subscription ids
    :return:
    """
    subs = []
    if len(sub_ids) <= openapi_threads:
        threads_num = 1
    else:
        threads_num = openapi_threads

    chunks = get_subscriptions_chunks(sub_ids, len(sub_ids) / threads_num)
    threads = []
    results = {}
    for i in range(threads_num):
        t = threading.Thread(target=get_subscriptions_worker, args=[chunks[i], results, i])
        threads.append(t)
        t.start()

    for t in threads:
        t.join()

    for result in results:
        subs.extend(results[result])

    for sub in subs:
        for resource_type in sub.resource_types:
            resource_type.save(db_connection)

    return subs


def set_and_save_subscriptions(subs, sts, migration_type):
    """
    Checks whether this is CI, CS, Legacy VPS models or other.
    :param subs: list of Subscription objects
    :param sts: list of Service Template objects
    :return:
    """
    for sub in subs:
        for st in sts:
            if sub.st_id == st.st_id:
                sub.set_resource_model(st.resource_model)
        sub.set_migration_type(migration_type)
        sub.save(db_connection)


def get_status():
    """
    Gets subscriptions objects from migration DB and output ID-->migration_status
    :return:
    """
    subscriptions = get_all_subs()
    for subscription in subscriptions:
        log.info('|subscription_id|' + '\t' + '|migration_status|' + '\t' + '|oaci_status|' + '\t' + '|migration_type|')
        log.info('{}\t\t{}\t\t{}\t\t{}'.format(subscription.subscription_id, subscription.migration_status,
                                               subscription.oaci_status, subscription.migration_type))


def get_all_subs():
    """
    Gets all subscriptions from migration DB and transform output into Subscription Object
    :return:
    """
    subs = db_connection.get_all_subscription_from_db()
    subscriptions = []
    for sub in subs:
        try:
            subscription = Subscription.from_db(db_connection=db_connection, sub_id=sub[0])
        except:
            continue
        subscriptions.append(subscription)
    return subscriptions


def check_if_oaci_activated(customer_id, subscription_id):
    """
    Calls Instance Manager for VEs of particular Customer and Subscription
    In case subscription has VEs -> OACI marked as active, status 2
    In case no VEs found but has active usage for 3 months -> OACI marked as was-active, status 1
    In case no VEs found -> OACI marked as not-activated for that Subscription, status 0
    :param customer_id: Customer id
    :param subscription_id: Subscription id
    :return:
    """
    ves = get_customer_ves(paci=paci, customer_id=customer_id, subscription_id=subscription_id)
    usage = get_usage_last_3_month(paci=paci, customer_id=customer_id, subscription_id=subscription_id)
    if bool(ves):
        db_connection.update_subscription_oaci_status(sub_id=subscription_id, status='active')
    elif bool(usage):
        db_connection.update_subscription_oaci_status(sub_id=subscription_id, status='was_active')
    else:
        db_connection.update_subscription_oaci_status(sub_id=subscription_id, status='inactive')


def inject_resources_worker(chunk, results, thread_id):
    """
    Sync Resource Types for specified Subscriptions in Thread
    :param chunk: chunk from the List of Subscriptions ids
    :param results: Dict, where each thread will store results
    :param thread_id: id of the Thread
    :return:
    """
    subs = []

    for sub_id in chunk:
        limits = []
        for resource_id, resource_name, limit in default_cs_resources_tuple:
            limits.append({'resource_type_id': resource_id, 'limit': limit})
        try:
            set_resource_type_limits(sub_id, limits)
        except Exception as e:
            log.err(traceback.format_exc())
            log.err(
                "[injection] cannot inject resources into subscription {}, most likely wrong resource {}".format(sub_id,
                                                                                                                 e))
            continue

        subs.append(sub_id)

    if len(subs) > 0:
        results[thread_id] = subs


def inject_resources_into_subscriptions(sub_ids):
    """
    Sync Resource Types for specified Subscriptions
    :param sub_ids: List of Subscriptions ids
    :return:
    """
    subs = []

    if len(sub_ids) <= openapi_threads:
        threads_num = 1
    else:
        threads_num = openapi_threads

    results = {}
    chunks = get_subscriptions_chunks(sub_ids, len(sub_ids) / threads_num)
    threads = []
    for i in range(threads_num):
        t = threading.Thread(target=inject_resources_worker, args=[chunks[i], results, i])
        threads.append(t)
        t.start()

    for t in threads:
        t.join()

    for result in results:
        subs.extend(results[result])

    for sub_id in subs:
        db_connection.update_subscription_status(sub_id=sub_id, status='injected')


def finalize_migration_worker(chunk, results, thread_id):
    """
    Finalize CI->CS migration, turn off old reporting, sync with Service Plan, in Thread
    :param chunk: List of chunk of Subscriptions ids
    :param results: Dict, where each thread will store results
    :param thread_id: id of the Thread
    :return:
    """
    subs = []

    t_pba_api = init_pba()

    for subscription in chunk:
        try:
            sync_with_plan(pba_api=t_pba_api, subscription_id=subscription.subscription_id,
                           default_cs_resources_tuple=default_cs_resources_tuple)
        except Exception as e:
            log.err(traceback.format_exc())
            log.err("[finalize] failed to add resource rates for subscription {}, "
                    "please check subscription and Service Plan and rerun the stage, {}"
                    .format(subscription.subscription_id, e))
            continue

        subs.append(subscription.subscription_id)
        log.debug(
            "[finalize] all resource rates for subscription {} has been sync-ed".format(subscription.subscription_id))

    if len(subs) > 0:
        results[thread_id] = subs


def finalize_migration(sub_ids):
    """
    Finalize CI->CS migration, sync with Service Plan
    :param sub_ids: List of Subscriptions ids
    :return:
    """
    subs = []

    if len(sub_ids) <= pba_api_threads:
        threads_num = 1
    else:
        threads_num = pba_api_threads

    chunks = get_subscriptions_chunks(sub_ids, len(sub_ids) / threads_num)
    threads = []
    results = {}

    oa_con = uSysDB.connect()

    for i in range(threads_num):
        t = threading.Thread(target=finalize_migration_worker, args=[chunks[i], results, i])
        threads.append(t)
        t.start()

    for t in threads:
        t.join()

    for result in results:
        subs.extend(results[result])

    for sub_id in subs:
        try:
            turnoff_ci_ui(subscription_id=sub_id, con=oa_con)
        except Exception as e:
            log.err(traceback.format_exc())
            log.err("[finalize] failed turn off old UI for subscription {},"
                    "please check subscription and Service Plan and rerun the stage, {}".format(sub_id, e))
            subs.remove(sub_id)
            continue
        log.debug("[finalize] all old UI has been turned off for subscription {} ".format(sub_id))

    for sub_id in subs:
        db_connection.update_subscription_status(sub_id=sub_id, status='migrated')


def turnoff_ui_only(sub_ids):
    """
     Finalize CI->CS migration, turn off old UI only
    :param sub_ids: List of Subscriptions ids
    :return:
    """
    oa_con = uSysDB.connect()
    for sub_id in sub_ids:
        try:
            turnoff_ci_ui(subscription_id=sub_id.subscription_id, con=oa_con)
        except Exception as e:
            log.err("[fast_turnoff_ui] failed turn off old UI for subscription {},"
                    "please check subscription and Service Plan and rerun the stage".format(sub_id.subscription_id, e))
            sub_ids.remove(sub_id)
            continue

        log.debug(
            "[fast_turnoff_ui] all old UI has been turned off for subscription {} ".format(sub_id.subscription_id))

    for sub_id in sub_ids:
        db_connection.update_subscription_status(sub_id=sub_id.subscription_id, status='evaluated')


def turnoff_ci_ui(subscription_id, con):
    """
    Removes limit for c2u.ve resource for subscription in OA DB
    :param con: live connection to OA Sys DB
    :param subscription_id: Subscription id
    :return:
    """
    cur = con.cursor()
    cur.execute("SELECT rt_instance_id FROM subs_resources sr LEFT JOIN "
                "resource_types rt ON (sr.rt_id = rt.rt_id) LEFT JOIN "
                "resource_classes rc ON (rt.class_id = rc.class_id) WHERE "
                "rc.name = 'c2u.ve' AND sr.sub_id = {}".format(subscription_id))
    rt_instance_id = cur.fetchone()[0]
    cur.execute("UPDATE subs_resources SET "
                "sub_limit = 0, curr_usage = 0, own_usage = 0 WHERE "
                "rt_instance_id = {}".format(rt_instance_id))
    con.commit()
